;****** HUGE ***************************************************
;****** Switches RM to huge real mode                   ********
;***************************************************************
; Seg limits of all data seg regs are set to max.
; Must be safe:
; 1. Checks processor state before PM switch.
; 2. Does not use non-const global variables.
;***************************************************************
                ideal
                %crefref
                %noincl
                %nomacs
                p386

;		model	large
;               include "help.asi"
                include "386.asi"	;for Flags_ZF
;               include "seg0.asi"
;               include "system.asi"



;###### CODE ###################################################
;               codeseg
        segment ASM_TEXT word public 'CODE'
                assume  cs:ASM_TEXT


;***************************************************************
;****** Sub:    _is_rm32
;***************************************************************
;****** Entry:  none
;****** Return: AX=1 - it is a 386+ in real mode
;******         AX=0 - otherwise
;****** Uses:   Flags
;***************************************************************
        global  _is_rm32:far
        proc    _is_rm32 far

;****** Check processor ****************************************
                push    sp
                pop     ax
                cmp     ax,sp
                jne     @@ret   ;ZF=0, it is a 86/186, not a 386+

                smsw    ax
                test    al,1	;MSW_PE
                jnz     @@ret   ;ZF=0, we're in PM/VM, not RM

                push    ds dx
                push    0
                pop     ds
                mov     ax,offset excp6
		mov	dx,cs
                pushf
                cli
                xchg    [06h*4  ],ax
                xchg    [06h*4+2],dx
               ;cmp     sp,sp           ;force ZF=1 - already done
                xchg    eax,eax         ;triggers excp6 on 286 - sets ZF=0
                mov     [06h*4  ],ax
                mov     [06h*4+2],dx
                popf
                pop     dx ds

;****** Return *************************************************
@@ret:          mov     ax,1    ;ZF=1 - AX=1
                jz      @@skip
                dec     ax      ;ZF=0 - AX=0
@@skip:         ret
        endp    _is_rm32


;***************************************************************
;****** Sub:    excp6
;***************************************************************
;****** Entry:  excp 6 trap
;****** Return: ZF=0; IP += 2;
;****** Uses:   None
;***************************************************************
        proc    excp6   far
oldFlags        =       (word bp+6)
oldCS           =       (word bp+4)
oldIP           =       (word bp+2)
oldBP           =       (word bp+0)

;****** Init ***************************************************
                push	bp
		mov	bp,sp

;****** Update stack *******************************************
                add     [oldIP],2
                and     [byte low oldFlags],not Flags_ZF

;****** Return *************************************************
@@ret:          pop	bp
                iret
        endp    excp6


;***************************************************************
;****** Sub:    _huge_rm
;***************************************************************
;****** Entry:  none
;****** Return: AX=1 - CPU is in huge RM now
;******         AX=0 - error: not a 386+ in RM
;****** Uses:   EAX, Flags
;***************************************************************
        global  _huge_rm:far
        proc    _huge_rm far
                local   GDTR    :pword, \
                        oldGDTR :pword  = TEMP_SIZE

;****** Init ***************************************************
                call    _is_rm32
                test    ax,ax
                jnz     @@rm32
                ret

@@rm32:         enter   TEMP_SIZE,0
                push    es ds fs gs ecx edx
                sgdt    [oldGDTR]

;****** Load GDTR **********************************************
                mov     edx,cs          ;yes, it zero extends!
                shl     edx,4
                mov     cx,offset GDT   ;Coulda be simple add edx,ofs GDT
                movzx   ecx,cx          ;but TLINK will moan...
                add     edx,ecx
                mov     [word GDTR],0FFFFh      ;GDT limit
                mov     [dword GDTR+2],edx      ;GDT base
                lgdt    [GDTR]

;****** Jump into PM, load all data sregs **********************
                mov     edx,cr0
                or      dl,01h          ;CR0_PE on
                mov     cx,0008h
                mov     ax,ss           ;save SS

                pushf
                cli
                mov     cr0,edx
                jmp     short $+2       ;*Required*!
                                        ;3+ NOPs also work fine (chkd on 386)
                mov     es,cx   ;Lim = 4Gb
                mov     ss,cx   ;
                mov     ds,cx   ;
                mov     fs,cx   ;
                mov     gs,cx   ;

;****** Return to RM *******************************************
                and     dl,0feh         ;CR0_PE off
                mov     cr0,edx         ;Limits are *not* reset to 64Kb!
                jmp     short $+2
                mov     ss,ax           ;restore SS
                popf

;****** Return *************************************************
@@ret:          mov     ax,1            ;success
                lgdt    [oldGDTR]
                pop     edx ecx gs fs ds es
                leave
                ret

;****** Const data *********************************************
                org     $-8     ;save 8 bytes
;selector 0000:
GDT             dd      ?,?

;selector 0008: Data seg [0,FFFFFFFF]
                ;       lim_lo              base_lo
                dw      1111111111111111b,  0000000000000000b
                db      00000000b,10010010b,10001111b,00000000b
                ;       base_med  PPlSEDWA  GD  l_hi  base_hi

        endp    _huge_rm

        ends    ASM_TEXT

        end

;###### END OF FILE ############################################
